import math
from statistics import mode

import numpy as np
import pandas as pd

from start_time_estimator.concurrency_oracle import DirectlyFollowsConcurrencyOracle, AlphaConcurrencyOracle, \
    HeuristicsConcurrencyOracle, DeactivatedConcurrencyOracle, OverlappingConcurrencyOracle
from start_time_estimator.config import ConcurrencyOracleType, ReEstimationMethod, ResourceAvailabilityType, OutlierStatistic, Configuration
from start_time_estimator.resource_availability import SimpleResourceAvailability, CalendarResourceAvailability


class StartTimeEstimator:
    def __init__(self, event_log: pd.DataFrame, config: Configuration):
        # Set event log
        self.event_log = event_log
        # Set configuration
        self.config = config
        # Set log IDs to ease access within class
        self.log_ids = config.log_ids
        # Set concurrency oracle
        if self.config.concurrency_oracle_type == ConcurrencyOracleType.DEACTIVATED:
            self.concurrency_oracle = DeactivatedConcurrencyOracle(self.config)
        elif self.config.concurrency_oracle_type == ConcurrencyOracleType.DF:
            self.concurrency_oracle = DirectlyFollowsConcurrencyOracle(self.event_log, self.config)
        elif self.config.concurrency_oracle_type == ConcurrencyOracleType.ALPHA:
            self.concurrency_oracle = AlphaConcurrencyOracle(self.event_log, self.config)
        elif self.config.concurrency_oracle_type == ConcurrencyOracleType.HEURISTICS:
            self.concurrency_oracle = HeuristicsConcurrencyOracle(self.event_log, self.config)
        elif self.config.concurrency_oracle_type == ConcurrencyOracleType.OVERLAPPING:
            self.concurrency_oracle = OverlappingConcurrencyOracle(self.event_log, self.config)
        else:
            raise ValueError("No concurrency oracle defined!")
        # Set resource availability
        if self.config.resource_availability_type == ResourceAvailabilityType.SIMPLE:
            self.resource_availability = SimpleResourceAvailability(self.event_log, self.config)
        elif self.config.resource_availability_type == ResourceAvailabilityType.WITH_CALENDAR:
            self.resource_availability = CalendarResourceAvailability(self.event_log, self.config)
        else:
            raise ValueError("No resource availability defined!")

    def estimate(self, replace_recorded_start_times: bool = False) -> pd.DataFrame:
        """
        Estimate the start times of each activity instance in the event log based on the resource availability and enablement times with
        the configuration defined in the parameters.

        :param replace_recorded_start_times:    If 'true', replace the start time column with the estimated start
                                                times, if 'false', the estimation is placed in its own column.

        :return: A copy of the event log with the estimated start time, the resource availability time, and the enablement time for each
        activity instance.
        """
        # Copy self event log to allow lunching this method many times
        event_log = self.event_log.copy()
        # Compute resource availability time if not already in the log
        if self.log_ids.available_time not in event_log.columns:
            self.resource_availability.add_resource_availability_times(event_log)
        # Compute enablement time if not already in the log
        if self.log_ids.enabled_time not in event_log.columns:
            self.concurrency_oracle.add_enabled_times(event_log)
        # Assign estimated start timestamps
        event_log[self.log_ids.estimated_start_time] = event_log[
            [self.log_ids.available_time, self.log_ids.enabled_time]
        ].max(axis=1, skipna=True, numeric_only=False)
        # Reuse current start times as estimation if the option is enabled
        if self.config.reuse_current_start_times:
            event_log.loc[
                ~pd.isna(event_log[self.log_ids.start_time]),
                self.log_ids.estimated_start_time
            ] = event_log[self.log_ids.start_time]
        # Re-estimate as instant those activities declared as instant
        event_log.loc[
            event_log[self.log_ids.activity].isin(self.config.instant_activities),
            self.log_ids.estimated_start_time
        ] = event_log[self.log_ids.end_time]
        # Re-estimate start time of those events with an estimated duration over the threshold
        if not math.isnan(self.config.outlier_threshold):
            self._re_estimate_durations_over_threshold(event_log)
        # Fix start time of those events for which it could not be estimated (with pd.NaT)
        if self.config.re_estimation_method == ReEstimationMethod.SET_INSTANT:
            self._set_instant_non_estimated_start_times(event_log)
        else:
            self._re_estimate_non_estimated_start_times(event_log)
        # If replacement to true, set estimated as start times
        if replace_recorded_start_times:
            event_log[self.log_ids.start_time] = event_log[self.log_ids.estimated_start_time]
            event_log.drop([self.log_ids.estimated_start_time], axis=1, inplace=True)
        # Return estimated event log
        return event_log

    def _re_estimate_durations_over_threshold(self, event_log: pd.DataFrame):
        # Get only events with estimated start time
        estimated_events = event_log[~pd.isna(event_log[self.log_ids.estimated_start_time])]
        # For each event, if the duration is over the threshold, set the defined statistic
        for activity, events in estimated_events.groupby(self.log_ids.activity):
            statistic_duration = self._apply_statistic(events[self.log_ids.end_time] - events[self.log_ids.estimated_start_time])
            duration_limit = self.config.outlier_threshold * statistic_duration
            event_log.loc[
                (event_log[self.log_ids.activity] == activity) &
                (~pd.isna(event_log[self.log_ids.estimated_start_time])) &
                ((event_log[self.log_ids.end_time] - event_log[self.log_ids.estimated_start_time]) > duration_limit),
                self.log_ids.estimated_start_time
            ] = event_log[self.log_ids.end_time] - duration_limit

    def _set_instant_non_estimated_start_times(self, event_log: pd.DataFrame):
        # Identify events with non_estimated as start time
        # and set their duration to instant
        event_log.loc[
            pd.isna(event_log[self.log_ids.estimated_start_time]),
            self.log_ids.estimated_start_time
        ] = event_log[self.log_ids.end_time]

    def _re_estimate_non_estimated_start_times(self, event_log: pd.DataFrame):
        # Get only events with estimated start time
        estimated_events = event_log[~pd.isna(event_log[self.log_ids.estimated_start_time])]
        # For each event, if the duration is over the threshold, set the defined statistic
        for activity, events in estimated_events.groupby(self.log_ids.activity):
            durations = (events[self.log_ids.end_time] - events[self.log_ids.estimated_start_time]).values
            statistic_duration = self._get_activity_duration(durations)
            event_log.loc[
                (event_log[self.log_ids.activity] == activity) &
                pd.isna(event_log[self.log_ids.estimated_start_time]),
                self.log_ids.estimated_start_time
            ] = event_log[self.log_ids.end_time] - statistic_duration
        # Set remaining non estimated activity instances to instant (those of activities with no estimated time)
        event_log[self.log_ids.estimated_start_time].fillna(event_log[self.log_ids.end_time], inplace=True)

    def _get_activity_duration(self, durations):
        if self.config.re_estimation_method == ReEstimationMethod.MODE:
            return mode(durations)
        elif self.config.re_estimation_method == ReEstimationMethod.MEDIAN:
            return np.median(durations)
        elif self.config.re_estimation_method == ReEstimationMethod.MEAN:
            return np.mean(durations)
        else:
            raise ValueError("Unselected re-estimation method for events with non-estimated start time!")

    def _apply_statistic(self, durations):
        if self.config.outlier_statistic == OutlierStatistic.MODE:
            return mode(durations)
        elif self.config.outlier_statistic == OutlierStatistic.MEDIAN:
            return np.median(durations)
        elif self.config.outlier_statistic == OutlierStatistic.MEAN:
            return np.mean(durations)
        else:
            raise ValueError("Unselected outlier statistic for events with estimated duration over the established!")



import enum
from dataclasses import dataclass, field

from pix_framework.log_ids import DEFAULT_CSV_IDS, EventLogIDs


class ReEstimationMethod(enum.Enum):
    SET_INSTANT = 1
    MODE = 2
    MEDIAN = 3
    MEAN = 4


class OutlierStatistic(enum.Enum):
    MODE = 1
    MEDIAN = 2
    MEAN = 3


class ConcurrencyOracleType(enum.Enum):
    DEACTIVATED = 1
    DF = 2
    ALPHA = 3
    HEURISTICS = 4
    OVERLAPPING = 5


class ResourceAvailabilityType(enum.Enum):
    SIMPLE = 1  # Consider all the events that each resource performs
    WITH_CALENDAR = 2  # Future possibility considering also the resource calendars and non-working days


@dataclass
class ConcurrencyThresholds:
    df: float = 0.9
    l2l: float = 0.9
    l1l: float = 0.9


@dataclass
class Configuration:
    """Class storing the configuration parameters for the start time estimation.

    Attributes:
        log_ids                     Identifiers for each key element (e.g. executed activity or resource).
        concurrency_oracle_type     Concurrency oracle to use (e.g. heuristics miner's concurrency oracle).
        resource_availability_type  Resource availability engine to use (e.g. using resource calendars).
        missing_resource            String to identify the events with missing resource (it is avoided in
                                    the resource availability calculation).
        re_estimation_method        Method (e.g. median) to re-estimate the start times that couldn't be
                                    estimated due to lack of resource availability and causal predecessors.
        bot_resources               Set of resource IDs corresponding bots, in order to set their events as
                                    instant.
        instant_activities          Set of instantaneous activities, in order to set their events as instant.
        concurrency_thresholds      Thresholds for the concurrency oracle. The three thresholds [df], [l1l],
                                    and [l2l] are used in the Heuristics oracle. In the overlapping oracle,
                                    only [df] is used.
        reuse_current_start_times   Do not estimate the start times of those activities with already recorded
                                    start time (caution, the instant activities and bot resources will still
                                    be set as instant).
        consider_start_times        Consider start times when checking for the enabled time of an activity in
                                    the concurrency oracle, if 'true', do not consider the events which end
                                    time is after the start time of the current activity instance, they overlap
                                    so no causality between them. In the case of the resource availability, if
                                    'true', search the availability as the previous end before the start of the
                                    current activity, not its end.
        outlier_statistic           Statistic (e.g. median) to calculate the most typical duration from the
                                    distribution of each activity durations to consider and re-estimate the
                                    outlier events which estimated duration is higher.
        outlier_threshold           Threshold to control outliers, those events with estimated durations over
        working_schedules           Dictionary with the resources as key and the working calendars (RCalendar)
                                    as value.
    """
    log_ids: EventLogIDs = field(default_factory=lambda: DEFAULT_CSV_IDS)
    concurrency_oracle_type: ConcurrencyOracleType = ConcurrencyOracleType.HEURISTICS
    resource_availability_type: ResourceAvailabilityType = ResourceAvailabilityType.SIMPLE
    missing_resource: str = "NOT_SET"
    re_estimation_method: ReEstimationMethod = ReEstimationMethod.MEDIAN
    bot_resources: set = field(default_factory=set)
    instant_activities: set = field(default_factory=set)
    concurrency_thresholds: ConcurrencyThresholds = field(default_factory=lambda: ConcurrencyThresholds())
    reuse_current_start_times: bool = False
    consider_start_times: bool = False
    outlier_statistic: OutlierStatistic = OutlierStatistic.MEDIAN
    outlier_threshold: float = float('nan')
    working_schedules: dict = field(default_factory=dict)



from collections import Counter
from typing import Optional

import pandas as pd

from start_time_estimator.config import Configuration
from start_time_estimator.utils import zip_with_next
from pix_framework.log_ids import EventLogIDs


class ConcurrencyOracle:
    def __init__(self, concurrency: dict, config: Configuration):
        # Dict with the concurrency: self.concurrency[A] = set of activities concurrent with A
        self.concurrency = concurrency
        # Configuration parameters
        self.config = config
        # Set log IDs to ease access within class
        self.log_ids = config.log_ids

    def enabled_since(self, trace, event) -> pd.Timestamp:
        # Get enabling activity instance or NA if none
        enabling_activity_instance = self.enabling_activity_instance(trace, event)
        # Return enabling activity instance
        return enabling_activity_instance[self.log_ids.end_time] if isinstance(enabling_activity_instance, pd.Series) else pd.NaT

    def enabling_activity_instance(self, trace, event) -> Optional[pd.Series]:
        # Get the list of previous end times
        previous_end_times = trace[
            (trace[self.log_ids.end_time] < event[self.log_ids.end_time]) &  # i) previous to the current one;
            ((not self.config.consider_start_times) or  # ii) if parallel check is activated,
             (trace[self.log_ids.end_time] <= event[self.log_ids.start_time])) &  # not overlapping;
            (~trace[self.log_ids.activity].isin(self.concurrency[event[self.log_ids.activity]]))  # iii) with no concurrency;
            ][self.log_ids.end_time]
        # Get enabling activity instance or NA if none
        enabling_activity_instance = trace.loc[previous_end_times.idxmax()] if len(previous_end_times) > 0 else None
        # Return enabling activity instance
        return enabling_activity_instance

    def add_enabled_times(self, event_log: pd.DataFrame, set_nat_to_first_event: bool = False, include_enabling_activity: bool = False):
        """
        Add the enabled time of each activity instance to the received event log based on the concurrency relations established in the
        class instance (extracted from the event log passed to the instantiation). For the first event on each trace, set the start of the
        trace as value.

        :param event_log:                   event log to add the enabled time information to.
        :param set_nat_to_first_event:      if False, use the start of the trace as enabled time for the activity instances with no previous
                                            activity enabling them, otherwise use pd.NaT.
        :param include_enabling_activity:   if True, add a column with the label of the activity enabling the current one.
        """
        # For each trace in the log, estimate the enabled time of its events
        indexes, enabled_times, enabling_activities = [], [], []
        for (case_id, trace) in event_log.groupby(self.log_ids.case):
            # Compute trace start time
            if self.log_ids.start_time in trace:
                trace_start_time = min(trace[self.log_ids.start_time].min(), trace[self.log_ids.end_time].min())
            else:
                trace_start_time = trace[self.log_ids.end_time].min()
            # Get the enabling activity of each event
            for index, event in trace.iterrows():
                indexes += [index]
                enabling_activity_instance = self.enabling_activity_instance(trace, event)
                # Store enabled time
                if isinstance(enabling_activity_instance, pd.Series):
                    # Use computed value
                    enabling_activity_label = enabling_activity_instance[self.log_ids.activity]
                    enabled_times += [enabling_activity_instance[self.log_ids.end_time]]
                else:
                    # No enabling activity, use trace start or NA
                    enabling_activity_label = pd.NA
                    enabled_times += [pd.NaT] if set_nat_to_first_event else [trace_start_time]
                # Store enabled activity label if necessary
                if include_enabling_activity:
                    enabling_activities += [enabling_activity_label]
        # Set all trace enabled times (and enabling activities if necessary) at once
        if include_enabling_activity:
            event_log.loc[indexes, self.log_ids.enabling_activity] = enabling_activities
        event_log.loc[indexes, self.log_ids.enabled_time] = enabled_times
        event_log[self.log_ids.enabled_time] = pd.to_datetime(event_log[self.log_ids.enabled_time], utc=True)


class DeactivatedConcurrencyOracle(ConcurrencyOracle):
    def __init__(self, config: Configuration):
        # Super (with empty concurrency)
        super(DeactivatedConcurrencyOracle, self).__init__({}, config)

    def enabled_since(self, trace, event) -> pd.Timestamp:
        return pd.NaT

    def enabling_activity_instance(self, trace, event) -> Optional[pd.Series]:
        return None


class DirectlyFollowsConcurrencyOracle(ConcurrencyOracle):
    def __init__(self, event_log: pd.DataFrame, config):
        # Default with no concurrency (all directly-follows relations)
        activities = event_log[config.log_ids.activity].unique()
        concurrency = {activity: set() for activity in activities}
        # Super
        super(DirectlyFollowsConcurrencyOracle, self).__init__(concurrency, config)


class AlphaConcurrencyOracle(ConcurrencyOracle):
    def __init__(self, event_log: pd.DataFrame, config: Configuration):
        # Alpha concurrency
        # Initialize dictionary for directly-follows relations df_relations[A][B] = number of times B following A
        df_relations = _get_df_relations(event_log, config.log_ids)
        # Create concurrency if there is a directly-follows relation in both directions
        concurrency = {}
        activities = event_log[config.log_ids.activity].unique()
        for act_a in activities:
            concurrency[act_a] = set()
            for act_b in activities:
                if act_a != act_b and act_a in df_relations.get(act_b, []) and act_b in df_relations.get(act_a, []):
                    # Concurrency relation AB, add it to A
                    concurrency[act_a].add(act_b)
        # Super
        super(AlphaConcurrencyOracle, self).__init__(concurrency, config)


def _get_df_relations(event_log: pd.DataFrame, log_ids: EventLogIDs) -> dict:
    # Initialize dictionary for directly-follows relations df_relations[A][B] = number of times B following A
    df_relations = {activity: {} for activity in event_log[log_ids.activity].unique()}
    # Fill dictionary with directly-follows relations
    for (key, trace) in event_log.groupby(log_ids.case):
        for (i, current_event), (j, future_event) in zip_with_next(trace.iterrows()):
            current_activity = current_event[log_ids.activity]
            future_activity = future_event[log_ids.activity]
            # Store df relation
            df_relations[current_activity][future_activity] = df_relations[current_activity].get(future_activity, 0) + 1
    return df_relations


class HeuristicsConcurrencyOracle(ConcurrencyOracle):
    def __init__(self, event_log: pd.DataFrame, config: Configuration):
        # Heuristics concurrency
        activities = event_log[config.log_ids.activity].unique()
        # Get matrices for:
        # - Directly-follows relations: df_count[A][B] = number of times B following A
        # - Directly-follows dependency values: df_dependency[A][B] = value of certainty that there is a df-relation between A and B
        # - Length-2 loop values: l2l_dependency[A][B] = value of certainty that there is a l2l relation between A and B (A-B-A)
        (df_count, df_dependency, l2l_dependency) = _get_heuristics_matrices(event_log, activities, config)
        # Create concurrency if there is a directly-follows relation in both directions
        concurrency = {}
        for act_a in activities:
            concurrency[act_a] = set()
            for act_b in activities:
                if (act_a != act_b and  # They are not the same activity
                        df_count[act_a].get(act_b, 0) > 0 and  # 'B' follows 'A' at least once
                        df_count[act_b].get(act_a, 0) > 0 and  # 'A' follows 'B' at least once
                        l2l_dependency[act_a].get(act_b, 0) < config.concurrency_thresholds.l2l and  # 'A' and 'B' are not a length 2 loop
                        abs(df_dependency[act_a].get(act_b, 0)) < config.concurrency_thresholds.df):  # The df relations are weak
                    # Concurrency relation AB, add it to A
                    concurrency[act_a].add(act_b)
        # Super
        super(HeuristicsConcurrencyOracle, self).__init__(concurrency, config)


def _get_heuristics_matrices(event_log: pd.DataFrame, activities: list, config: Configuration) -> (dict, dict, dict):
    # Initialize dictionary for directly-follows relations df_count[A][B] = number of times B following A
    df_count = {activity: {} for activity in activities}
    # Initialize dictionary for length 2 loops
    l2l_count = {activity: {} for activity in activities}
    # Count directly-follows and l2l relations
    for (key, trace) in event_log.groupby(config.log_ids.case):
        previous_activity = None
        # Iterate the events of the trace in pairs: (e1, e2), (e2, e3), (e3, e4)...
        for (i, current_event), (j, future_event) in zip_with_next(trace.iterrows()):
            current_activity = current_event[config.log_ids.activity]
            future_activity = future_event[config.log_ids.activity]
            # Store df relation
            df_count[current_activity][future_activity] = df_count[current_activity].get(future_activity, 0) + 1
            # Process l2l
            if previous_activity:
                # Increase value if there is a length 2 loop (A-B-A)
                if previous_activity == future_activity:
                    l2l_count[previous_activity][current_activity] = l2l_count[previous_activity].get(current_activity, 0) + 1
            # Save previous activity
            previous_activity = current_activity
    # Fill df and l1l dependency matrices
    df_dependency = {activity: {} for activity in activities}
    l1l_dependency = {activity: 0 for activity in activities}
    for act_a in activities:
        for act_b in activities:
            if act_a != act_b:
                # Process directly follows dependency value A -> B
                ab = df_count[act_a].get(act_b, 0)
                ba = df_count[act_b].get(act_a, 0)
                df_dependency[act_a][act_b] = (ab - ba) / (ab + ba + 1)
            else:
                # Process length 1 loop value
                aa = df_count[act_a].get(act_a, 0)
                l1l_dependency[act_a] = aa / (aa + 1)
    # Fill l2l dependency matrix
    l2l_dependency = {activity: {} for activity in activities}
    for act_a in activities:
        for act_b in activities:
            if act_a != act_b and \
                    l1l_dependency[act_a] < config.concurrency_thresholds.l1l and \
                    l1l_dependency[act_b] < config.concurrency_thresholds.l1l:
                # Process directly follows dependency value A -> B
                aba = l2l_count[act_a].get(act_b, 0)
                bab = l2l_count[act_b].get(act_a, 0)
                l2l_dependency[act_a][act_b] = (aba + bab) / (aba + bab + 1)
            else:
                l2l_dependency[act_a][act_b] = 0
    # Return matrices with dependency values
    return df_count, df_dependency, l2l_dependency


class OverlappingConcurrencyOracle(ConcurrencyOracle):
    def __init__(self, event_log: pd.DataFrame, config: Configuration):
        # Get the activity labels
        activities = set(event_log[config.log_ids.activity])
        # Get matrix with the frequency of each activity happening overlapping with the rest and in directly-follows order
        overlapping_relations = _get_overlapping_matrix(event_log, activities, config)
        # Create concurrency if the overlapping relations is higher than the threshold specifies
        concurrency = {activity: set() for activity in activities}
        already_checked = set()
        for act_a in activities:
            # Store as already checked to avoid redundant checks
            already_checked.add(act_a)
            # Get the number of occurrences of A per case
            occurrences_a = Counter(event_log[event_log[config.log_ids.activity] == act_a][config.log_ids.case])
            for act_b in (activities - already_checked):
                # Get the number of occurrences of B per case
                occurrences_b = Counter(event_log[event_log[config.log_ids.activity] == act_b][config.log_ids.case])
                # Compute number of times they co-occur
                co_occurrences = sum([
                    occurrences_a[case_id] * occurrences_b[case_id]
                    for case_id
                    in set(list(occurrences_a.keys()) + list(occurrences_b.keys()))
                ])
                # Check if the proportion of overlapping occurrences is higher than the established threshold
                if co_occurrences > 0:
                    overlapping_ratio = overlapping_relations[act_a].get(act_b, 0) / co_occurrences
                    if overlapping_ratio >= config.concurrency_thresholds.df:
                        # Concurrency relation AB, add it
                        concurrency[act_a].add(act_b)
                        concurrency[act_b].add(act_a)
        # Set flag to consider start times also when individually checking enabled time
        config.consider_start_times = True
        # Super
        super(OverlappingConcurrencyOracle, self).__init__(concurrency, config)


def _get_overlapping_matrix(event_log: pd.DataFrame, activities: set, config: Configuration) -> dict:
    # Initialize dictionary for overlapping relations df_count[A][B] = number of times B overlaps with A
    overlapping_relations = {activity: {} for activity in activities}
    # Count overlapping relations
    for (key, trace) in event_log.groupby(config.log_ids.case):
        # Iterate the events of the trace
        for (i, event) in trace.iterrows():
            current_activity = event[config.log_ids.activity]
            # Get labels of overlapping activity instances
            overlapping_labels = trace[
                ((event[config.log_ids.start_time] < trace[config.log_ids.start_time]) &  # The current event starts while the other
                 (trace[config.log_ids.start_time] < event[config.log_ids.end_time])) |  # is being executed; OR
                ((event[config.log_ids.start_time] < trace[config.log_ids.end_time]) &  # the current event ends while the other
                 (trace[config.log_ids.end_time] < event[config.log_ids.end_time])) |  # is being executed; OR
                ((trace[config.log_ids.start_time] <= event[config.log_ids.start_time]) &  # the other event starts and
                 (event[config.log_ids.end_time] <= trace[config.log_ids.end_time]) &  # ends within the current one,
                 (event[config.log_ids.activity] != trace[config.log_ids.activity]))  # and it's not the current one.
                ][config.log_ids.activity]
            for overlapping_activity in overlapping_labels:
                overlapping_relations[current_activity][overlapping_activity] = (
                        overlapping_relations[current_activity].get(overlapping_activity, 0) + 1
                )
    # Return matrix with dependency values
    return overlapping_relations



from datetime import datetime

import pandas as pd
from pix_framework.calendar.availability import get_last_available_timestamp

from start_time_estimator.config import Configuration


class ResourceAvailability:
    def __init__(self, performed_events: dict, working_schedules: dict, config: Configuration):
        # Store dictionary with the resources as key and all its events as value
        self.performed_events = performed_events
        # Store dictionary with the resources as key and all its events as value
        self.working_schedules = working_schedules
        # Configuration parameters
        self.config = config
        # Set log IDs to ease access within class
        self.log_ids = config.log_ids

    def available_since(self, resource: str, event) -> datetime:
        # Get the end timestamp of the previous activity instance executed by the same resource
        if resource == self.config.missing_resource:
            # If the resource is missing return pd.NaT
            timestamp_previous_event = pd.NaT
        elif resource in self.config.bot_resources:
            # If the resource has been marked as 'bot resource', return the same timestamp
            timestamp_previous_event = event[self.log_ids.end_time]
        else:
            # If existing resource, take the latest timestamp previous to [timestamp]
            resource_events = self.performed_events[resource]
            timestamp_previous_event = resource_events.where(
                (resource_events < event[self.log_ids.end_time]) &
                ((not self.config.consider_start_times) or (resource_events <= event[self.log_ids.start_time]))
            ).max()
            if pd.isna(timestamp_previous_event):
                timestamp_previous_event = pd.NaT
            # If there are non-working periods from the latest previous
            # end timestamp, take the end of the last non-working period
            if resource in self.working_schedules:
                activity_start = event[self.log_ids.start_time] if self.config.consider_start_times else event[self.log_ids.end_time]
                timestamp_previous_event = get_last_available_timestamp(
                    start=timestamp_previous_event,
                    end=activity_start,
                    schedule=self.working_schedules[resource]
                )
        # Return previous timestamp where the resource became available
        return timestamp_previous_event

    def add_resource_availability_times(self, event_log: pd.DataFrame):
        """
        Add the resource availability time of each activity instance to the received event log. For the first event of each resource, set
        pd.NaT.

        :param event_log: event log to add the resource availability time information to.
        """
        # For each trace in the log, estimate the enabled time of its events
        indexes = []
        resource_availability_times = []
        for (case_id, trace) in event_log.groupby(self.log_ids.case):
            # Get the resource availability times
            for index, event in trace.iterrows():
                indexes += [index]
                resource_availability_time = self.available_since(event[self.log_ids.resource], event)
                resource_availability_times += [resource_availability_time]
        # Set all enabled times at once
        event_log.loc[indexes, self.log_ids.available_time] = resource_availability_times
        event_log[self.log_ids.available_time] = pd.to_datetime(event_log[self.log_ids.available_time], utc=True)


class SimpleResourceAvailability(ResourceAvailability):
    def __init__(self, event_log: pd.DataFrame, config: Configuration):
        # Create a dictionary with the resources as key and all its end events as value
        resources = {str(i) for i in event_log[config.log_ids.resource].unique()}
        resources_calendar = {}
        for resource in (resources - config.bot_resources):
            resources_calendar[resource] = event_log[event_log[config.log_ids.resource] == resource][config.log_ids.end_time]
        # Super
        super(SimpleResourceAvailability, self).__init__(resources_calendar, {}, config)


class CalendarResourceAvailability(ResourceAvailability):
    def __init__(self, event_log: pd.DataFrame, config: Configuration):
        # Create a dictionary with the resources as key and all its end events as value
        resources = {str(i) for i in event_log[config.log_ids.resource].unique()}
        resources_calendar = {}
        for resource in (resources - config.bot_resources):
            resources_calendar[resource] = event_log[event_log[config.log_ids.resource] == resource][config.log_ids.end_time]
        # Super
        super(CalendarResourceAvailability, self).__init__(resources_calendar, config.working_schedules, config)
